from pydantic import BaseModel
from typing import Dict, List, Optional

from dasl_api import (
    WorkspaceV1ExportConfig,
    WorkspaceV1ExportConfigSlackConfig,
    WorkspaceV1ExportConfigWebhookConfig,
    WorkspaceV1ExportConfigWebhookConfigDestination,
    WorkspaceV1WorkspaceConfig,
    WorkspaceV1WorkspaceConfigSpec,
    WorkspaceV1WorkspaceConfigSpecComputeGroupLimitsInner,
    WorkspaceV1WorkspaceConfigSpecDefaultConfig,
    WorkspaceV1WorkspaceConfigSpecDetectionRuleMetadata,
    WorkspaceV1WorkspaceConfigSpecManagedRetentionInner,
    WorkspaceV1WorkspaceConfigSpecManagedRetentionInnerOverridesInner,
    WorkspaceV1WorkspaceConfigSpecObservables,
    WorkspaceV1WorkspaceConfigSpecObservablesKindsInner,
    WorkspaceV1WorkspaceConfigSpecSystemTablesConfig,
    WorkspaceV1DefaultConfig,
    WorkspaceV1DefaultConfigComputeGroupOverridesValue,
)

from .helpers import Helpers
from .types import Metadata, ResourceStatus


class ExportConfig(BaseModel):
    """
    Configuration settings for exporting notables, operational alerts, and
    other events.

    Attributes:
        destination (Optional[str]):
        export_open_only (Optional[bool]): If true, only open events are
            exported. If false, any event not in state . ClosedAsExported
            will be exported.
        webhook_config (Optional[ExportConfig.WebhookConfig]): Set when
            exporting to a webhook.
        slack_config (Optional[ExportConfig.SlackConfig]): Set when exporting
            to slack.
    """

    class WebhookDestination(BaseModel):
        """
        Configuration settings for exporting to a webhook.

        Attributes:
            value (Optional[str]):
            scope (Optional[str]):
            key (Optional[str]):
        """

        value: Optional[str] = None
        scope: Optional[str] = None
        key: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[WorkspaceV1ExportConfigWebhookConfigDestination],
        ) -> "ExportConfig.WebhookDestination":
            if obj is None:
                return None
            return ExportConfig.WebhookDestination(
                value=obj.value,
                scope=obj.scope,
                key=obj.key,
            )

        def to_api_obj(self) -> WorkspaceV1ExportConfigWebhookConfigDestination:
            return WorkspaceV1ExportConfigWebhookConfigDestination(
                value=self.value,
                scope=self.scope,
                key=self.key,
            )

    class WebhookConfig(BaseModel):
        """
        Configuration settings for exporting to a webhook.

        Attributes:
            destination (Optional[ExportConfig.WebhookDestination]):
        """

        destination: Optional["ExportConfig.WebhookDestination"] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[WorkspaceV1ExportConfigWebhookConfig],
        ) -> "ExportConfig.WebhookConfig":
            if obj is None:
                return None
            return ExportConfig.WebhookConfig(
                destination=ExportConfig.WebhookDestination.from_api_obj(
                    obj.destination
                )
            )

        def to_api_obj(self) -> WorkspaceV1ExportConfigWebhookConfig:
            if self.destination is None:
                return None
            return WorkspaceV1ExportConfigWebhookConfig(
                destination=self.destination.to_api_obj(),
            )

    class SlackConfig(BaseModel):
        """
        Configuration settings for exporting to Slack.

        Attributes:
            token (Optional[ExportConfig.WebhookDestination]):
            channel (Optional[str]):
            message (Optional[str]):
        """

        token: Optional["ExportConfig.WebhookDestination"] = None
        channel: Optional[str] = None
        message: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[WorkspaceV1ExportConfigSlackConfig],
        ) -> "ExportConfig.SlackConfig":
            if obj is None:
                return None
            return ExportConfig.SlackConfig(
                token=ExportConfig.WebhookDestination.from_api_obj(obj.token),
                channel=obj.channel,
                message=obj.message,
            )

        def to_api_obj(self) -> WorkspaceV1ExportConfigSlackConfig:
            token = None
            if self.token is not None:
                token = self.token.to_api_obj()
            return WorkspaceV1ExportConfigSlackConfig(
                token=token,
                channel=self.channel,
                message=self.message,
            )

    destination: Optional[str] = None
    export_open_only: Optional[bool] = None
    webhook_config: Optional["ExportConfig.WebhookConfig"] = None
    slack_config: Optional["ExportConfig.SlackConfig"] = None

    @staticmethod
    def from_api_obj(obj: Optional[WorkspaceV1ExportConfig]) -> "ExportConfig":
        if obj is None:
            return None
        return ExportConfig(
            destination=obj.destination,
            export_open_only=obj.export_open_only,
            webhook_config=ExportConfig.WebhookConfig.from_api_obj(obj.webhook_config),
            slack_config=ExportConfig.SlackConfig.from_api_obj(obj.slack_config),
        )

    def to_api_obj(self) -> WorkspaceV1ExportConfig:
        slack_config = None
        if self.slack_config is not None:
            slack_config = self.slack_config.to_api_obj()

        webhook_config = None
        if self.webhook_config is not None:
            webhook_config = self.webhook_config.to_api_obj()

        return WorkspaceV1ExportConfig(
            destination=self.destination,
            export_open_only=self.export_open_only,
            webhook_config=webhook_config,
            slack_config=slack_config,
        )


class WorkspaceConfigObservables(BaseModel):
    """
    Configuration of observables for the workspace.

    Attributes:
        kinds (Optional[List[ObservablesKinds]]):
        relationships (Optional[List[str]]):
    """

    class ObservablesKinds(BaseModel):
        """
        Configuration of a concrete kind of observable.

        Attributes:
            name (Optional[str]):
            sql_type (Optional[str]):
        """

        name: Optional[str] = None
        sql_type: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: WorkspaceV1WorkspaceConfigSpecObservablesKindsInner,
        ) -> "WorkspaceConfigObservables.ObservablesKinds":
            return WorkspaceConfigObservables.ObservablesKinds(
                name=obj.name,
                sql_type=obj.sql_type,
            )

        def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecObservablesKindsInner:
            return WorkspaceV1WorkspaceConfigSpecObservablesKindsInner(
                name=self.name,
                sql_type=self.sql_type,
            )

    kinds: Optional[List[ObservablesKinds]] = None
    relationships: Optional[List[str]] = None

    @staticmethod
    def from_api_obj(
        obj: Optional[WorkspaceV1WorkspaceConfigSpecObservables],
    ) -> "WorkspaceConfigObservables":
        if obj is None:
            return None

        kinds = None
        if obj.kinds is not None:
            kinds = [
                WorkspaceConfigObservables.ObservablesKinds.from_api_obj(item)
                for item in obj.kinds
            ]
        return WorkspaceConfigObservables(
            kinds=kinds,
            relationships=obj.relationships,
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecObservables:
        kinds = None
        if self.kinds is not None:
            kinds = [item.to_api_obj() for item in self.kinds]
        return WorkspaceV1WorkspaceConfigSpecObservables(
            kinds=kinds,
            relationships=self.relationships,
        )


class ComputeGroupLimits(BaseModel):
    """
    Configuration of limits for the number of resources that can be placed
    in a user-defined compute group.

    Attributes:
        group_name (Optional[str]): Name of the compute group for this limit.
        limit (Optional[int]): Limit of maximum number of resources in the group.
    """

    group_name: Optional[str] = None
    limit: Optional[int] = None

    @staticmethod
    def from_api_obj(
        obj: WorkspaceV1WorkspaceConfigSpecComputeGroupLimitsInner,
    ) -> "ComputeGroupLimits":
        return ComputeGroupLimits(
            group_name=obj.group_name,
            limit=obj.limit,
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecComputeGroupLimitsInner:
        return WorkspaceV1WorkspaceConfigSpecComputeGroupLimitsInner(
            group_name=self.group_name,
            limit=self.limit,
        )


class DefaultConfig(BaseModel):
    """
    Configuration of the schemas, notebook storage locations, checkpoint
    storage locations, and so forth, for each concrete resource type and
    a global fallback that applies to resources which do not have a
    specified DefaultConfig.

    Attributes:
        datasources (Optional[DefaultConfig.Config]): Configuration that
            applies to DataSources. May be omitted.
        transforms (Optional[DefaultConfig.Config]): Configuration that
            applies to Transforms. May be omitted.
        rules (Optional[DefaultConfig.Config]): Configuration that applies
            to Rules. May be omitted.
        var_global (Optional[DefaultConfig.Config]): Configuration that
            applies globally to resources without a resource-specific
            configuration specified. Must be specified.
    """

    class Config(BaseModel):
        """
        Default configuration for a specific resource type.

        Attributes:
            notebook_location (Optional[str]): A location for user created/
                edited/provided notebook.
            bronze_schema (Optional[str]): Name of the bronze schema in
                the catalog.
            silver_schema (Optional[str]): Name of the silver schema in
                the catalog.
            gold_schema (Optional[str]): Name of the gold schema in
                the catalog.
            catalog_name (Optional[str]): The catalog name to use as the
                resource's default.
            default_max_resources_per_job (Optional[int]): Default maximum
                number of resources that can be placed into a single job,
                subject to compute_group_overrides.
            checkpoint_location (Optional[str]): The location to store
                checkpoints for streaming writes. If not provided, the
                daslStoragePath will be used.
            compute_group_overrides (Optional[Dict[str, DefaultConfig.
                Config.ComputeGroupOverrides]]): Overrides for the maximum
                number of resources that can be placed into a single job,
                keyed by the compute group name.
        """

        class ComputeGroupOverrides(BaseModel):
            """
            Specifies the maximum number of resources that can be placed
            into a single job.

            Attributes:
                max_resources_per_job (Optional[int]): The maximum number
                    of resources that can be placed into a job before a
                    new job will be created.
            """

            max_resources_per_job: Optional[int] = None

            @staticmethod
            def from_api_obj(
                obj: Optional["DefaultConfig.Config.ComputeGroupOverrides"],
            ) -> "DefaultConfig.Config.ComputeGroupOverrides":
                return DefaultConfig.Config.ComputeGroupOverrides(
                    max_resources_per_job=obj.max_resources_per_job,
                )

            def to_api_obj(self) -> WorkspaceV1DefaultConfigComputeGroupOverridesValue:
                return WorkspaceV1DefaultConfigComputeGroupOverridesValue(
                    max_resources_per_job=self.max_resources_per_job,
                )

        notebook_location: Optional[str] = None
        bronze_schema: Optional[str] = None
        silver_schema: Optional[str] = None
        gold_schema: Optional[str] = None
        catalog_name: Optional[str] = None
        default_max_resources_per_job: Optional[int] = None
        checkpoint_location: Optional[str] = None
        compute_group_overrides: Optional[
            Dict[str, "DefaultConfig.Config.ComputeGroupOverrides"]
        ] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[WorkspaceV1DefaultConfig],
        ) -> "DefaultConfig.Config":
            if obj is None:
                return None

            compute_group_overrides = None
            if obj.compute_group_overrides is not None:
                compute_group_overrides = {
                    key: DefaultConfig.Config.ComputeGroupOverrides.from_api_obj(value)
                    for key, value in obj.compute_group_overrides.items()
                }
            return DefaultConfig.Config(
                notebook_location=obj.notebook_location,
                bronze_schema=obj.bronze_schema,
                silver_schema=obj.silver_schema,
                gold_schema=obj.gold_schema,
                catalog_name=obj.catalog_name,
                default_max_resources_per_job=obj.default_max_resources_per_job,
                checkpoint_location=obj.checkpoint_location,
                compute_group_overrides=compute_group_overrides,
            )

        def to_api_obj(self) -> WorkspaceV1DefaultConfig:
            compute_group_overrides = None
            if self.compute_group_overrides is not None:
                compute_group_overrides = {
                    key: value.to_api_obj()
                    for key, value in self.compute_group_overrides.items()
                }
            return WorkspaceV1DefaultConfig(
                notebook_location=self.notebook_location,
                bronze_schema=self.bronze_schema,
                silver_schema=self.silver_schema,
                gold_schema=self.gold_schema,
                catalog_name=self.catalog_name,
                default_max_resources_per_job=self.default_max_resources_per_job,
                checkpoint_location=self.checkpoint_location,
                compute_group_overrides=compute_group_overrides,
            )

    datasources: Optional["DefaultConfig.Config"] = None
    transforms: Optional["DefaultConfig.Config"] = None
    rules: Optional["DefaultConfig.Config"] = None
    var_global: Optional["DefaultConfig.Config"] = None

    @staticmethod
    def from_api_obj(
        obj: WorkspaceV1WorkspaceConfigSpecDefaultConfig,
    ) -> "DefaultConfig":
        return DefaultConfig(
            datasources=DefaultConfig.Config.from_api_obj(obj.datasources),
            transforms=DefaultConfig.Config.from_api_obj(obj.transforms),
            rules=DefaultConfig.Config.from_api_obj(obj.rules),
            var_global=DefaultConfig.Config.from_api_obj(obj.var_global),
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecDefaultConfig:
        datasources = None
        if self.datasources is not None:
            datasources = self.datasources.to_api_obj()

        transforms = None
        if self.transforms is not None:
            transforms = self.transforms.to_api_obj()

        rules = None
        if self.rules is not None:
            rules = self.rules.to_api_obj()

        var_global = None
        if self.var_global is not None:
            var_global = self.var_global.to_api_obj()

        return WorkspaceV1WorkspaceConfigSpecDefaultConfig(
            datasources=datasources,
            transforms=transforms,
            rules=rules,
            var_global=var_global,
        )


class ManagedRetention(BaseModel):
    """
    Configuration of cleanup jobs for old data stored by DASL.

    Attributes:
        catalog (str): The name of the catalog.
        var_schema (str): The name of the schema.
        column (Optional[str]):
        duration (Optional[str]): The duration for which to retain data in
            the schema. It may also be "forever" to indicate data should not
            be deleted.
        overrides (Optional[List[ManagedRetention.Overrides]]): Overrides for
            per-table retention rules within the specified catalog and schema.
    """

    class Overrides(BaseModel):
        """
        Per-table retention override settings.

        Attributes:
            table (str): The name of the table in the catalog and schema.
            column (Optional[str]):
            duration (Optional[str]): The duration for which to retain data
                in the table. It may also be "forever" to indicate data
                should not be deleted.
        """

        table: str
        column: Optional[str] = None
        duration: Optional[str] = None  # TODO: consider using duration type?

        @staticmethod
        def from_api_obj(
            obj: WorkspaceV1WorkspaceConfigSpecManagedRetentionInnerOverridesInner,
        ) -> "ManagedRetention.Overrides":
            return ManagedRetention.Overrides(
                table=obj.table,
                column=obj.column,
                duration=obj.duration,
            )

        def to_api_obj(
            self,
        ) -> WorkspaceV1WorkspaceConfigSpecManagedRetentionInnerOverridesInner:
            return WorkspaceV1WorkspaceConfigSpecManagedRetentionInnerOverridesInner(
                table=self.table,
                column=self.column,
                duration=self.duration,
            )

    catalog: str
    var_schema: str
    column: Optional[str] = None
    duration: Optional[str] = None
    overrides: Optional[List["ManagedRetention.Overrides"]] = None

    @staticmethod
    def from_api_obj(
        obj: WorkspaceV1WorkspaceConfigSpecManagedRetentionInner,
    ) -> "ManagedRetention":
        overrides = None
        if obj.overrides is not None:
            overrides = [
                ManagedRetention.Overrides.from_api_obj(item) for item in obj.overrides
            ]
        return ManagedRetention(
            catalog=obj.catalog,
            var_schema=obj.var_schema,
            column=obj.column,
            duration=obj.duration,
            overrides=overrides,
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecManagedRetentionInner:
        overrides = None
        if self.overrides is not None:
            overrides = [item.to_api_obj() for item in self.overrides]
        return WorkspaceV1WorkspaceConfigSpecManagedRetentionInner(
            catalog=self.catalog,
            var_schema=self.var_schema,
            column=self.column,
            duration=self.duration,
            overrides=overrides,
        )


class SystemTablesConfig(BaseModel):
    """
    Specifies where system data (i.e. metadata managed by the DASL
    control plane) will be stored.

    Attributes:
        catalog_name (str): The name of the catalog in which to store data.
        var_schema (str): The name of the schema in which to create tables.
    """

    catalog_name: str
    var_schema: str

    @staticmethod
    def from_api_obj(
        obj: WorkspaceV1WorkspaceConfigSpecSystemTablesConfig,
    ) -> "SystemTablesConfig":
        return SystemTablesConfig(
            catalog_name=obj.catalog_name,
            var_schema=obj.var_schema,
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecSystemTablesConfig:
        return WorkspaceV1WorkspaceConfigSpecSystemTablesConfig(
            catalog_name=self.catalog_name,
            var_schema=self.var_schema,
        )


class DetectionRuleMetadata(BaseModel):
    """
    DetectionRuleMetadata

    Attributes:
        detection_categories (Optional[List[str]]):
    """

    detection_categories: Optional[List[str]] = None

    @staticmethod
    def from_api_obj(
        obj: Optional[WorkspaceV1WorkspaceConfigSpecDetectionRuleMetadata],
    ) -> "DetectionRuleMetadata":
        if obj is None:
            return None
        return DetectionRuleMetadata(
            detection_categories=obj.detection_categories,
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfigSpecDetectionRuleMetadata:
        return WorkspaceV1WorkspaceConfigSpecDetectionRuleMetadata(
            detection_categories=self.detection_categories,
        )


class WorkspaceConfig(BaseModel):
    """
    General configuration settings for the Workspace.

    Attributes:
        metadata (Optional[Metadata]): Common resource metadata; generally
            managed by the control plane.
        system_tables_config (SystemTablesConfig): Configuration for the
            storage of metadata by the control plane.
        default_sql_warehouse (Optional[str]): Default SQL warehouse to use
            for executing certain queries. May be overridden in some cases.
        detection_rule_metadata (Optional[DetectionRuleMetadata]):
        notable_export (Optional[ExportConfig]): Settings related to the
            export of notables to various destinations.
        operational_alert_export (Optional[ExportConfig]): Settings related
            to the export of operational alerts to various destinations.
        observables (Optional[WorkspaceConfigObservables]): Declaration of
            the types of observables generated by the system.
        compute_group_limits (Optional[List[ComputeGroupLimits]]): Limits
            on the number of resources placed in a manually specified
            compute group - corresponds to the number of resources that
            are grouped into a single Job.
        dasl_storage_path (Optional[str]): The path to a directory where
            DASL can store internal files and state.
        default_config (Optional[DefaultConfig]): Configuration settings
            regarding storage of bronze, silver, and gold tables and
            related assets for each resource type.
        managed_retention (Optional[List[ManagedRetention]]): Configuration
            of regular cleanup (i.e. pruning) jobs for various catalogs,
            schemas, and tables.
        status (Optional[ResourceStatus]): Common resource status; wholly
            managed by the control plane.
    """

    metadata: Optional[Metadata] = None
    system_tables_config: SystemTablesConfig
    default_sql_warehouse: Optional[str] = None
    detection_rule_metadata: Optional[DetectionRuleMetadata] = None
    notable_export: Optional[ExportConfig] = None
    operational_alert_export: Optional[ExportConfig] = None
    observables: Optional[WorkspaceConfigObservables] = None
    compute_group_limits: Optional[List[ComputeGroupLimits]] = None
    dasl_storage_path: Optional[str] = None
    default_config: Optional[DefaultConfig] = None
    managed_retention: Optional[List[ManagedRetention]] = None
    status: Optional[ResourceStatus] = None

    @staticmethod
    def from_api_obj(obj: WorkspaceV1WorkspaceConfig) -> "WorkspaceConfig":
        spec = obj.spec

        compute_group_limits = None
        if spec.compute_group_limits is not None:
            compute_group_limits = [
                ComputeGroupLimits.from_api_obj(item)
                for item in spec.compute_group_limits
            ]

        managed_retention = None
        if spec.managed_retention is not None:
            managed_retention = [
                ManagedRetention.from_api_obj(item) for item in spec.managed_retention
            ]

        return WorkspaceConfig(
            metadata=Metadata.from_api_obj(obj.metadata),
            system_tables_config=SystemTablesConfig.from_api_obj(
                spec.system_tables_config
            ),
            default_sql_warehouse=spec.default_sql_warehouse,
            detection_rule_metadata=DetectionRuleMetadata.from_api_obj(
                spec.detection_rule_metadata
            ),
            notable_export=ExportConfig.from_api_obj(spec.notable_export),
            operational_alert_export=ExportConfig.from_api_obj(
                spec.operational_alert_export
            ),
            observables=WorkspaceConfigObservables.from_api_obj(spec.observables),
            compute_group_limits=compute_group_limits,
            dasl_storage_path=spec.dasl_storage_path,
            default_config=DefaultConfig.from_api_obj(spec.default_config),
            managed_retention=managed_retention,
            status=ResourceStatus.from_api_obj(obj.status),
        )

    def to_api_obj(self) -> WorkspaceV1WorkspaceConfig:
        compute_group_limits = None
        if self.compute_group_limits is not None:
            compute_group_limits = [
                item.to_api_obj() for item in self.compute_group_limits
            ]

        managed_retention = None
        if self.managed_retention is not None:
            managed_retention = [item.to_api_obj() for item in self.managed_retention]

        to_api_obj = lambda o: o.to_api_obj()

        return WorkspaceV1WorkspaceConfig(
            api_version="v1",
            kind="WorkspaceConfig",
            metadata=Helpers.maybe(to_api_obj, self.metadata),
            spec=WorkspaceV1WorkspaceConfigSpec(
                system_tables_config=self.system_tables_config.to_api_obj(),
                default_sql_warehouse=self.default_sql_warehouse,
                detection_rule_metadata=Helpers.maybe(
                    to_api_obj, self.detection_rule_metadata
                ),
                notable_export=Helpers.maybe(to_api_obj, self.notable_export),
                operational_alert_export=Helpers.maybe(
                    to_api_obj, self.operational_alert_export
                ),
                observables=Helpers.maybe(to_api_obj, self.observables),
                compute_group_limits=compute_group_limits,
                dasl_storage_path=self.dasl_storage_path,
                default_config=Helpers.maybe(to_api_obj, self.default_config),
                managed_retention=managed_retention,
            ),
            status=Helpers.maybe(to_api_obj, self.status),
        )



import os
import pytest
from urllib.parse import urlparse

from dasl_client import *

dasl_host = os.environ["DASL_API_URL"]
databricks_host = os.environ["DATABRICKS_HOST"]
databricks_client_id = os.environ["DATABRICKS_CLIENT_ID"]
databricks_client_secret = os.environ["DATABRICKS_CLIENT_SECRET"]
workspace = urlparse(databricks_host).hostname
app_client_id = "22853b93-68ba-4ae2-8e41-976417f501dd"
alternate_app_client_id = "335ac0d3-e0ea-4732-ba93-0277423b5029"


@pytest.fixture(scope="session")
def api_client():
    client = Client.new_workspace(
        admin_email="test@antimatter.io",
        app_client_id=app_client_id,
        service_principal_id=databricks_client_id,
        service_principal_secret=databricks_client_secret,
        workspace_url=databricks_host,
        dasl_host=dasl_host,
    )
    yield client


def test_admin_config(api_client):
    base_admin_config = AdminConfig(
        workspace_url=databricks_host,
        app_client_id=app_client_id,
        service_principal_id=databricks_client_id,
        service_principal_secret="********",
    )

    ac = api_client.get_admin_config()
    assert ac == base_admin_config

    other = AdminConfig(
        workspace_url=databricks_host,
        app_client_id=alternate_app_client_id,
        service_principal_id=databricks_client_id,
        service_principal_secret=databricks_client_secret,
    )
    api_client.put_admin_config(other)

    assert api_client.get_admin_config() == AdminConfig(
        workspace_url=databricks_host,
        app_client_id=alternate_app_client_id,
        service_principal_id=databricks_client_id,
        service_principal_secret="********",
    )

    ac.service_principal_secret = databricks_client_secret
    api_client.put_admin_config(ac)
    assert api_client.get_admin_config() == base_admin_config


def test_workspace_config(api_client):
    base_workspace_config = WorkspaceConfig(
        metadata=Metadata(
            name="config",
            workspace=workspace,
        ),
        dasl_storage_path="/Volumes/automated_test_cases/default/test",
        system_tables_config=SystemTablesConfig(
            catalog_name="automated_test_cases",
            var_schema="default",
        ),
        default_config=DefaultConfig(
            var_global=DefaultConfig.Config(
                bronze_schema="bronze",
                silver_schema="silver",
                gold_schema="gold",
                catalog_name="automated_test_cases",
            ),
        ),
    )

    api_client.put_config(base_workspace_config)
    got = api_client.get_config()

    # the server is going to populate created_timestamp, modified_timestamp,
    # version, and resource_status, so copy those over before comparing.
    base_workspace_config.metadata.created_timestamp = got.metadata.created_timestamp
    base_workspace_config.metadata.modified_timestamp = got.metadata.modified_timestamp
    base_workspace_config.metadata.version = got.metadata.version
    base_workspace_config.metadata.resource_status = got.metadata.resource_status

    assert api_client.get_config() == base_workspace_config

    base_workspace_config.default_config.var_global.bronze_schema = "bronze_new"
    api_client.put_config(base_workspace_config)
    got = api_client.get_config()
    base_workspace_config.metadata.modified_timestamp = got.metadata.modified_timestamp
    base_workspace_config.metadata.version = got.metadata.version
    base_workspace_config.metadata.resource_status = got.metadata.resource_status

    assert api_client.get_config() == base_workspace_config


def test_minimal_data_source(api_client):
    base_data_source = DataSource(
        source="test",
        schedule=Schedule(
            at_least_every="2h",
            enabled=True,
        ),
        bronze=BronzeSpec(
            bronze_table="test_bronze_table",
            skip_bronze_loading=False,
        ),
        silver=SilverSpec(),
        gold=GoldSpec(),
    )

    base_ds_1 = api_client.create_datasource("test_1", base_data_source)
    assert base_ds_1.source == base_data_source.source
    assert base_ds_1.schedule == base_data_source.schedule
    assert base_ds_1.bronze == base_data_source.bronze
    assert base_ds_1.silver == base_data_source.silver
    assert base_ds_1.gold == base_data_source.gold

    got = api_client.get_datasource("test_1")
    listed = []
    for ds in api_client.list_datasources():
        listed.append(ds)
    assert len(listed) == 1
    assert listed[0] == got

    # the server is going to populate created_timestamp, modified_timestamp,
    # version, and resource_status, so copy those over before comparing.
    base_ds_1.metadata.created_timestamp = got.metadata.created_timestamp
    base_ds_1.metadata.created_by = got.metadata.created_by
    base_ds_1.metadata.modified_timestamp = got.metadata.modified_timestamp
    base_ds_1.metadata.version = got.metadata.version
    base_ds_1.metadata.resource_status = got.metadata.resource_status
    assert api_client.get_datasource("test_1") == base_ds_1

    base_ds_2 = api_client.create_datasource("test_2", base_data_source)
    assert base_ds_2.source == base_data_source.source
    assert base_ds_2.schedule == base_data_source.schedule
    assert base_ds_2.bronze == base_data_source.bronze
    assert base_ds_2.silver == base_data_source.silver
    assert base_ds_2.gold == base_data_source.gold

    got_2 = api_client.get_datasource("test_2")
    listed = []
    for ds in api_client.list_datasources():
        listed.append(ds)
    assert len(listed) == 2
    assert listed[0] == got
    assert listed[1] == got_2

    base_ds_2.metadata.created_timestamp = got_2.metadata.created_timestamp
    base_ds_2.metadata.created_by = got_2.metadata.created_by
    base_ds_2.metadata.modified_timestamp = got_2.metadata.modified_timestamp
    base_ds_2.metadata.version = got_2.metadata.version
    base_ds_2.metadata.resource_status = got_2.metadata.resource_status
    assert api_client.get_datasource("test_2") == base_ds_2

    base_ds_2.bronze.bronze_table = "test_2"
    api_client.replace_datasource("test_2", base_ds_2)

    got_2 = api_client.get_datasource("test_2")
    base_ds_2.metadata.modified_timestamp = got_2.metadata.modified_timestamp
    base_ds_2.metadata.version = got_2.metadata.version
    base_ds_2.metadata.resource_status = got_2.metadata.resource_status

    assert api_client.get_datasource("test_2") == base_ds_2

    api_client.delete_datasource("test_1")
    listed = [
        item
        for item in api_client.list_datasources()
        if item.metadata.resource_status != "deletionPending"
    ]
    assert len(listed) == 1
    assert listed[0] == base_ds_2


def test_minimal_rule(api_client):
    base_rule = Rule(
        schedule=Schedule(
            at_least_every="2h",
            enabled=True,
        ),
        input=Rule.Input(
            stream=Rule.Input.Stream(
                tables=[
                    Rule.Input.Stream.Table(
                        name="test",
                    ),
                ],
            ),
        ),
        output=Rule.Output(
            summary="test",
        ),
    )

    base_rule_1 = api_client.create_rule("test_0", base_rule)
    assert base_rule_1.schedule == base_rule.schedule
    assert base_rule_1.input == base_rule.input
    assert base_rule_1.output == base_rule.output

    got = api_client.get_rule("test_0")
    listed = []
    for rule in api_client.list_rules():
        listed.append(rule)
    assert len(listed) == 1
    assert listed[0] == got

    # the server is going to populate created_timestamp, modified_timestamp,
    # version, and resource_status, so copy those over before comparing.
    base_rule_1.metadata.created_timestamp = got.metadata.created_timestamp
    base_rule_1.metadata.created_by = got.metadata.created_by
    base_rule_1.metadata.modified_timestamp = got.metadata.modified_timestamp
    base_rule_1.metadata.version = got.metadata.version
    base_rule_1.metadata.resource_status = got.metadata.resource_status
    assert api_client.get_rule("test_0") == base_rule_1

    base_rule_2 = api_client.create_rule("test_1", base_rule)
    assert base_rule_2.schedule == base_rule.schedule
    assert base_rule_2.input == base_rule.input
    assert base_rule_2.output == base_rule.output

    got_2 = api_client.get_rule("test_1")
    listed = []
    for rule in api_client.list_rules():
        listed.append(rule)
    assert len(listed) == 2
    assert listed[0] == got
    assert listed[1] == got_2

    base_rule_2.metadata.created_timestamp = got_2.metadata.created_timestamp
    base_rule_2.metadata.created_by = got_2.metadata.created_by
    base_rule_2.metadata.modified_timestamp = got_2.metadata.modified_timestamp
    base_rule_2.metadata.version = got_2.metadata.version
    base_rule_2.metadata.resource_status = got_2.metadata.resource_status
    assert api_client.get_rule("test_1") == base_rule_2

    base_rule_2.input.stream.tables[0].name = "test_1"
    api_client.replace_rule("test_1", base_rule_2)

    got_2 = api_client.get_rule("test_1")
    base_rule_2.metadata.modified_timestamp = got_2.metadata.modified_timestamp
    base_rule_2.metadata.version = got_2.metadata.version
    base_rule_2.metadata.resource_status = got_2.metadata.resource_status

    assert api_client.get_rule("test_1") == base_rule_2

    api_client.delete_rule("test_0")
    listed = [
        item
        for item in api_client.list_rules()
        if item.metadata.resource_status != "deletionPending"
    ]
    assert len(listed) == 1
    assert listed[0] == base_rule_2


def test_list_pagination(api_client):
    base_rule = Rule(
        schedule=Schedule(
            at_least_every="2h",
            enabled=True,
        ),
        input=Rule.Input(
            stream=Rule.Input.Stream(
                tables=[
                    Rule.Input.Stream.Table(
                        name="test",
                    ),
                ],
            ),
        ),
        output=Rule.Output(
            summary="test",
        ),
    )

    # create (remainder of) 10 rules for the test
    for i in range(8):
        api_client.create_rule(f"test_{i+2}", base_rule)

    # ensure all rules are returned for a list call with no params
    listed = []
    for rule in api_client.list_rules():
        listed.append(rule)
    assert len(listed) == 10

    for i in range(10):
        assert listed[i] == api_client.get_rule(f"test_{i}")

    # ensure the first 5 rules are returned when limit=5
    listed = []
    for rule in api_client.list_rules(limit=5):
        listed.append(rule)
    assert len(listed) == 5

    for i in range(5):
        assert listed[i] == api_client.get_rule(f"test_{i}")

    # ensure the last 5 rules are returned when limit=5, cursor=pagination_test_4
    listed = []
    for rule in api_client.list_rules(cursor="test_4", limit=5):
        listed.append(rule)
    assert len(listed) == 5

    for i in range(5):
        assert listed[i] == api_client.get_rule(f"test_{i+5}")

    # ensure the last 9 rules are returned when cursor=test_0
    listed = []
    for rule in api_client.list_rules(cursor="test_0"):
        listed.append(rule)
    assert len(listed) == 9

    for i in range(9):
        assert listed[i] == api_client.get_rule(f"test_{i+1}")



from typing import Optional, List, Dict
from pydantic import BaseModel

from dasl_api import (
    DbuiV1ObservableEventsList,
    DbuiV1ObservableEventsListItemsInnerNotable,
    DbuiV1ObservableEventsListItemsInner,
    DbuiV1TransformRequest,
    DbuiV1TransformRequestInput,
    DbuiV1TableColumnDetails,
    DbuiV1TransformRequestTransformsInner,
    DbuiV1TransformRequestTransformsInnerPresetOverrides,
    DbuiV1TransformResponse,
    DbuiV1TransformResponseStagesInner,
)

from .datasource import DataSource, FieldSpec, FieldUtils
from .helpers import Helpers


class Dbui(BaseModel):
    class TableColumnDetails(BaseModel):
        """
        Table column details.

        Attributes:
            name (Optional[str]): The column name.
            type_name (Optional[str]): The name of the column type.
            type_detail (Optional[str]): Additional information about the
                column's type.
            position (Optional[int]): The column's index in the table.
            nullable (Optional[bool]): Indicates if this column is nullable.
        """

        name: Optional[str] = None
        type_name: Optional[str] = None
        type_detail: Optional[str] = None
        position: Optional[int] = None
        nullable: Optional[bool] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[DbuiV1TableColumnDetails],
        ) -> Optional["Dbui.TableColumnDetails"]:
            if obj is None:
                return None
            return Dbui.TableColumnDetails(
                name=obj.name,
                type_name=obj.type_name,
                type_detail=obj.type_detail,
                position=obj.position,
                nullable=obj.nullable,
            )

        def to_api_obj(self) -> DbuiV1TableColumnDetails:
            return DbuiV1TableColumnDetails(
                name=self.name,
                type_name=self.type_name,
                type_detail=self.type_detail,
                position=self.position,
                nullable=self.nullable,
            )

    class ObservableEvents(BaseModel):
        class Notable(BaseModel):
            id: Optional[str] = None
            rule_name: Optional[str] = None
            summary: Optional[str] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[DbuiV1ObservableEventsListItemsInnerNotable],
            ) -> Optional["Dbui.ObservableEvents.Notable"]:
                if obj is None:
                    return None

                return Dbui.ObservableEvents.Notable(
                    id=obj.id, rule_name=obj.rule_name, summary=obj.summary
                )

        class Event(BaseModel):
            var_from: Optional[str] = None
            adjust_by: Optional[float] = None
            notable: Optional["Dbui.ObservableEvents.Notable"] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[DbuiV1ObservableEventsListItemsInner],
            ) -> Optional["Dbui.ObservableEvents.Event"]:
                if obj is None:
                    return None

                return Dbui.ObservableEvents.Event(
                    var_from=obj.var_from,
                    adjust_by=obj.adjust_by,
                    notable=Dbui.ObservableEvents.Notable.from_api_obj(obj.notable),
                )

        class EventsList(BaseModel):
            cursor: Optional[str] = None
            items: List["Dbui.ObservableEvents.Event"] = []

            @staticmethod
            def from_api_obj(
                obj: Optional[DbuiV1ObservableEventsList],
            ) -> Optional["Dbui.ObservableEvents.EventsList"]:
                if obj is None:
                    return None

                return Dbui.ObservableEvents.EventsList(
                    cursor=obj.cursor,
                    items=[
                        Dbui.ObservableEvents.Event.from_api_obj(item)
                        for item in obj.items
                    ],
                )


class TransformRequest(BaseModel):
    """
    The transform request identifies the starting data (either with an
    autoloader spec or an input block) and then specifies a chain of
    transforms to be performed on the data. The response includes the data
    at each intermediate stage (e.g. input/autoloaded data, pre-transform,
    silver).

    Attributes:
        input (TransformRequest.Input): The input block containing the
            columns metadata and data.
        autoloader_input (DataSource.Autoloader):
        use_preset (str): Indicates which preset to use for the transforms.
        transforms (List[TransformRequest.Transform]): A list of transform
            configurations.
    """

    class Input(BaseModel):
        """
        Input data for the transform request.

        Attributes:
            columns (List[Dbui.TableColumnDetails]): A list of metadata about
                the columns.
            data (List[Dict[str, str]]): The data represented as a list of
                dictionaries.
        """

        columns: List[Dbui.TableColumnDetails]
        data: List[Dict[str, str]]

        @staticmethod
        def from_api_obj(
            obj: Optional[DbuiV1TransformRequestInput],
        ) -> Optional["TransformRequest.Input"]:
            if obj is None:
                return None
            return TransformRequest.Input(
                columns=[
                    Dbui.TableColumnDetails.from_api_obj(item) for item in obj.columns
                ],
                data=obj.data,
            )

        def to_api_obj(self) -> DbuiV1TransformRequestInput:
            return DbuiV1TransformRequestInput(
                columns=[item.to_api_obj() for item in self.columns],
                data=self.data,
            )

    class Transform(BaseModel):
        """
        A transform configuration to apply to the data.

        Attributes:
            transform_type (str): The type of transform (one of
                SilverPreTransform, SilverTransform, Gold).
            use_pre_transform_preset (str): Indicates which pre-transform
                preset to use if requested.
            use_preset_table (str): Indicates which table to use within the
                preset's transform type for Silver and Gold.
            filter (str):
            post_filter (str):
            preset_overrides (TransformRequest.Transform.PresetOverrides):
                Overrides for the preset configuration.
            add_fields (List[FieldSpec]): Additional field specifications
                to add.
            utils (FieldUtils): Utility configurations for handling fields.
        """

        class PresetOverrides(BaseModel):
            """
            Preset overrides for a transform configuration.

            Attributes:
                omit_fields (List[str]): A list of fields to omit from
                    the preset.
            """

            omit_fields: Optional[List[str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[DbuiV1TransformRequestTransformsInnerPresetOverrides],
            ) -> Optional["TransformRequest.Transform.PresetOverrides"]:
                if obj is None:
                    return None
                return TransformRequest.Transform.PresetOverrides(
                    omit_fields=obj.omit_fields,
                )

            def to_api_obj(
                self,
            ) -> DbuiV1TransformRequestTransformsInnerPresetOverrides:
                return DbuiV1TransformRequestTransformsInnerPresetOverrides(
                    omit_fields=self.omit_fields,
                )

        transform_type: str
        use_pre_transform_preset: Optional[str] = None
        use_preset_table: Optional[str] = None
        filter: Optional[str] = None
        post_filter: Optional[str] = None
        preset_overrides: Optional["TransformRequest.Transform.PresetOverrides"] = None
        add_fields: Optional[List[FieldSpec]] = None
        utils: Optional[FieldUtils] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[DbuiV1TransformRequestTransformsInner],
        ) -> Optional["TransformRequest.Transform"]:
            if obj is None:
                return None
            add_fields = None
            if obj.add_fields is not None:
                add_fields = [FieldSpec.from_api_obj(item) for item in obj.add_fields]
            return TransformRequest.Transform(
                transform_type=obj.transform_type,
                use_pre_transform_preset=obj.use_pre_transform_preset,
                use_preset_table=obj.use_preset_table,
                filter=obj.filter,
                post_filter=obj.post_filter,
                preset_overrides=TransformRequest.Transform.PresetOverrides.from_api_obj(
                    obj.preset_overrides
                ),
                add_fields=add_fields,
                utils=FieldUtils.from_api_obj(obj.utils),
            )

        def to_api_obj(self) -> DbuiV1TransformRequestTransformsInner:
            add_fields = None
            if self.add_fields is not None:
                add_fields = [item.to_api_obj() for item in self.add_fields]
            to_api_obj = lambda o: o.to_api_obj()
            return DbuiV1TransformRequestTransformsInner(
                transform_type=self.transform_type,
                use_pre_transform_preset=self.use_pre_transform_preset,
                use_preset_table=self.use_preset_table,
                filter=self.filter,
                post_filter=self.post_filter,
                preset_overrides=Helpers.maybe(to_api_obj, self.preset_overrides),
                add_fields=add_fields,
                utils=Helpers.maybe(to_api_obj, self.utils),
            )

    input: Optional["TransformRequest.Input"] = None
    autoloader_input: Optional[DataSource.Autoloader] = None
    use_preset: Optional[str] = None
    transforms: List["TransformRequest.Transform"]

    @staticmethod
    def from_api_obj(obj: DbuiV1TransformRequest) -> "TransformRequest":
        return TransformRequest(
            input=TransformRequest.Input.from_api_obj(obj.input),
            autoloader_input=DataSource.Autoloader.from_api_obj(obj.autoloader_input),
            use_preset=obj.use_preset,
            transforms=[
                TransformRequest.Transform.from_api_obj(item) for item in obj.transforms
            ],
        )

    def to_api_obj(self) -> DbuiV1TransformRequest:
        to_api_obj = lambda o: o.to_api_obj()
        return DbuiV1TransformRequest(
            input=Helpers.maybe(to_api_obj, self.input),
            autoloader_input=Helpers.maybe(to_api_obj, self.autoloader_input),
            use_preset=self.use_preset,
            transforms=[item.to_api_obj() for item in self.transforms],
        )


class TransformResponse(BaseModel):
    """
    The transform response contains the results of the chain of transforms
    applied on the data.

    Attributes:
        stages (List[TransformResponse.Stages]): A list of stages
            representing each intermediate transform step.
    """

    class Stages(BaseModel):
        """
        A stage in the transform response.

        Attributes:
            transform_type (str): The type of transform applied in this stage
                (one of SilverPreTransform, SilverTransform, Gold, Input).
            columns (List[Dbui.TableColumnDetails]): A list of metadata about
                the columns returned in this stage.
            data (List[Dict[str, str]]): The data represented as a list
                of dictionaries.
        """

        transform_type: str
        columns: List[Dbui.TableColumnDetails]
        data: List[Dict[str, str]]

        @staticmethod
        def from_api_obj(
            obj: DbuiV1TransformResponseStagesInner,
        ) -> "TransformResponse.Stages":
            return TransformResponse.Stages(
                transform_type=obj.transform_type,
                columns=[
                    Dbui.TableColumnDetails.from_api_obj(item) for item in obj.columns
                ],
                data=obj.data,
            )

        def to_api_obj(self) -> DbuiV1TransformResponseStagesInner:
            return DbuiV1TransformResponseStagesInner(
                transform_type=self.transform_type,
                columns=[item.to_api_obj() for item in self.columns],
                data=self.data,
            )

    stages: List["TransformResponse.Stages"]

    @staticmethod
    def from_api_obj(obj: DbuiV1TransformResponse) -> "TransformResponse":
        return TransformResponse(
            stages=[TransformResponse.Stages.from_api_obj(item) for item in obj.stages],
        )

    def to_api_obj(self) -> DbuiV1TransformResponse:
        return DbuiV1TransformResponse(
            stages=[item.to_api_obj() for item in self.stages],
        )



import abc
import time
from datetime import datetime

from dasl_api import (
    api,
    WorkspaceV1AuthenticateRequest,
    ApiClient,
    WorkspaceV1RequestSecretRequest,
)
from databricks.sdk.errors import ResourceDoesNotExist

from dasl_client.conn.conn import get_base_conn
from dasl_client.errors.errors import handle_errors
from databricks.sdk import WorkspaceClient
from typing import Optional

# The minimum age of a conn that we allow. Generally, clients are issued for
# a small batch of operations (one or two requests), not long-running
# operations. This threshold should be tuned to the longest expected operation
# using a single conn.
EXPIRY_OVERLAP_SECONDS = 600


class Authorization(abc.ABC):
    """
    A common interface for Authentication
    """

    @abc.abstractmethod
    def client(self) -> ApiClient:
        raise NotImplementedError("conn method must be implemented")

    def workspace(self) -> str:
        raise NotImplementedError("client method must be implemented")


class ServiceAccountKeyAuth(Authorization):
    """
    Authorisation implementation for Service Account Keys
    """

    def __init__(
        self, workspace: str, service_account_key: str, host: Optional[str] = None
    ):
        self._workspace = workspace
        self._service_account_key = service_account_key
        self._client = get_base_conn(host=host)
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Service Account Key
        :return: The client name.
        """
        return self._workspace

    @handle_errors
    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        req = WorkspaceV1AuthenticateRequest(
            service_account_key=self._service_account_key
        )
        handler = api.WorkspaceV1Api(api_client=self._client)

        resp = handler.workspace_v1_authenticate(
            workspace=self._workspace, workspace_v1_authenticate_request=req
        )
        self._client.set_default_header("Authorization", f"Bearer {resp.token}")
        verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
        self.expiry = verification.expiry


class DatabricksTokenAuth(Authorization):
    """
    Authorization implementation using Databricks Tokens. Note that this will not work on most databricks
    workspaces using Premium or Enterprise tier databricks.
    """

    def __init__(self, workspace: str, token: str, host: Optional[str] = None):
        self._workspace = workspace
        self._databricks_token = token
        self._client = get_base_conn(host=host)
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Databricks Token
        :return: The client name.
        """
        return self._workspace

    @handle_errors
    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        req = WorkspaceV1AuthenticateRequest(
            databricks_api_token=self._databricks_token
        )
        handler = api.WorkspaceV1Api(api_client=self._client)

        resp = handler.workspace_v1_authenticate(
            workspace=self._workspace, workspace_v1_authenticate_request=req
        )
        self._client.set_default_header("Authorization", f"Bearer {resp.token}")
        verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
        self.expiry = verification.expiry


class DatabricksSecretAuth(Authorization):
    """
    Authorization implementation using Databricks Secrets
    """

    def __init__(self, workspace: str, host: Optional[str] = None):
        self._workspace = workspace
        self._client = get_base_conn(host=host)
        self._principal = WorkspaceClient().current_user.me().user_name
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Databricks Token
        :return: The client name.
        """
        return self._workspace

    @handle_errors
    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        # First we do a pre-authenticate call to refresh the secret. It doesn't really matter if we race
        # here with others, as long as the secret ends up with a recent value in it. The secret can be used
        # more than once, but it does expire
        req = WorkspaceV1RequestSecretRequest(
            principalName=self._principal,
        )
        handler = api.WorkspaceV1Api(api_client=self._client)
        resp = handler.workspace_v1_begin_authentication(
            workspace=self._workspace, workspace_v1_begin_authenticate_request=req
        )
        secret_name = resp.secret_name
        secret_value = ""
        for tries in range(3):
            try:
                secret_value = WorkspaceClient().secrets.get_secret(
                    secret_name, "token"
                )
                break
            except ResourceDoesNotExist:
                # Maybe there is a race here, let's retry
                time.sleep(0.5)
        if len(secret_value) == 0:
            raise RuntimeError(f"failed to complete secret auth")

        req = WorkspaceV1AuthenticateRequest(
            databricks_secret=secret_value,
        )
        handler = api.WorkspaceV1Api(api_client=self._client)

        resp = handler.workspace_v1_authenticate(
            workspace=self._workspace, workspace_v1_authenticate_request=req
        )
        self._client.set_default_header("Authorization", f"Bearer {resp.token}")
        verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
        self.expiry = verification.expiry



# setup.py

from setuptools import setup, find_packages

setup(
    name="dasl-client",
    version="0.0.0",
    author="Antimatter Team",
    author_email="support@antimatter.io",
    description="The DASL client library used for interacting with the DASL client.",
    long_description="TODO: Link to docs page or README.md.",
    long_description_content_type="text/markdown",
    url="https://github.com/antimatter/asl",
    packages=find_packages(),
    python_requires=">=3.8",
)



from copy import deepcopy
from typing import Any, Callable, Iterator, List, Optional, TypeVar

from dasl_api import (
    CoreV1Api,
    CoreV1DataSourceAutoloaderSpec,
    DbuiV1Api,
    WorkspaceV1Api,
    WorkspaceV1CreateWorkspaceRequest,
    api,
)
from pydantic import Field

from dasl_client.auth.auth import (
    Authorization,
    DatabricksTokenAuth,
    ServiceAccountKeyAuth,
)
from dasl_client.conn.conn import get_base_conn
from dasl_client.errors.errors import ConflictError, error_handler
from .helpers import Helpers

from .types import (
    AdminConfig,
    DataSource,
    Dbui,
    Metadata,
    Rule,
    WorkspaceConfig,
    TransformRequest,
    TransformResponse,
)


T = TypeVar("T")


class Client:
    """
    Antimatter security lakehouse client.
    """

    def __init__(
        self,
        auth: Authorization,
    ):
        """
        Initialize a new client. You should generally prefer to use
        the new_workspace function if creating a new workspace or the
        for_workspace function if connecting to an existing workspace.

        :param auth: Authorization instance for authorizing requests to
            the dasl control plane.
        :returns Client
        """
        self.auth = auth

    @staticmethod
    def new_workspace(
        admin_email: str,
        app_client_id: str,
        service_principal_id: str,
        service_principal_secret: str,
        workspace_url: Optional[str] = None,
        dasl_host: str = Helpers.default_dasl_host,
    ) -> "Client":
        """
        Register a new workspace and return a client for it.

        :param admin_email: The email address associated with the (DASL)
            workspace admin, if the workspace will be created.
        :param app_client_id: The Databricks app connection client ID
            to use for authentication calls related to the workspace.
        :param service_principal_id: The ID of the Databricks service
            principal that will interact with Databricks on your behalf.
        :param service_principal_secret: An OAuth secret that entitles
            the service principal to make Databricks API calls on your
            behalf.
        :param workspace_url: The full base URL of the Databricks workspace
            being registered. If you omit this value, it will be inferred
            if you are running within a Databricks notebook. Otherwise, an
            exception will be raised.
        :param dasl_host: The URL of the DASL server. This value should
            not generally be specified unless you are testing against
            an alternative environment.
        :returns Client
        """
        with error_handler():
            if workspace_url is None:
                workspace_url = Helpers.current_workspace_url()
            admin_config = AdminConfig(
                workspace_url=workspace_url,
                app_client_id=app_client_id,
                service_principal_id=service_principal_id,
                service_principal_secret=service_principal_secret,
            )
            workspace_name = Helpers.workspace_name_from_url(workspace_url)
            api_key = (
                api.WorkspaceV1Api(api_client=get_base_conn(host=dasl_host))
                .workspace_v1_create_workspace(
                    WorkspaceV1CreateWorkspaceRequest(
                        admin_user=admin_email,
                        workspace_name=workspace_name,
                        admin_config=admin_config.to_api_obj().spec,
                    )
                )
                .admin_service_account.apikey
            )
            return Client(
                ServiceAccountKeyAuth(workspace_name, api_key, host=dasl_host),
            )

    # TODO: should we allow user to use an API key here?
    @staticmethod
    def for_workspace(
        workspace_url: Optional[str] = None,
        service_account_token: Optional[str] = None,
        dasl_host: str = Helpers.default_dasl_host,
    ) -> "Client":
        """
        Create a client for the argument workspace, if specified, or
        the current workspace if running in databricks notebook context.

        :param workspace_url: The full base URL of the Databricks workspace
            being registered. If you omit this value, it will be inferred
            if you are running within a Databricks notebook. Otherwise,
            an exception will be raised.
        :param service_account_token: Antimatter service account token.
            If provided, the client will use this token for auth instead
            of (automatic) secret-based auth.
        :param dasl_host: The URL of the DASL server. This value should
            not generally be specified unless you are testing against
            an alternative environment.
        :returns Client
        """
        with error_handler():
            if workspace_url is None:
                workspace_url = Helpers.current_workspace_url()

            if service_account_token is None:
                return Client(
                    DatabricksSecretAuth(
                        Helpers.workspace_name_from_url(workspace_url),
                        host=dasl_host,
                    ),
                )
            else:
                return Client(
                    ServiceAccountKeyAuth(
                        Helpers.workspace_name_from_url(workspace_url),
                        service_account_token,
                        host=dasl_host,
                    )
                )

    @staticmethod
    def new_or_existing(
        admin_email: str,
        app_client_id: str,
        service_principal_id: str,
        service_principal_secret: str,
        workspace_url: Optional[str] = None,
        service_account_token: Optional[str] = None,
        dasl_host: str = Helpers.default_dasl_host,
    ) -> "Client":
        """
        Initialize a new client for the workspace associated with the
        argument Databricks workspace_url. If no such workspace exists,
        one will be created for you.

        :param admin_email: The email address associated with the (DASL)
            workspace admin, if the workspace will be created. Ignored if
            the workspace already exists.
        :param app_client_id: The Databricks app connection client ID
            to use for authentication calls related to the workspace. If
            the workspace already exists, the existing config will be
            updated to use this client ID.
        :param service_principal_id: The ID of the Databricks service
            principal that will interact with Databricks on your behalf.
            If the workspace already exists, the existing config will be
            updated to use this service principal ID.
        :param service_principal_secret: An OAuth secret that entitles
            the service principal to make Databricks API calls on your
            behalf. If the workspace already exists, the existing config
            will be updated to use this service principal secret.
        :param workspace_url: The full base URL of the Databricks workspace
            being registered. If you omit this value, it will be inferred
            if you are running within a Databricks notebook. Otherwise, an
            exception will be raised. If the workspace already exists, the
            existing config will be updated to use this value.
        :param service_account_token: Antimatter service account token.
            If provided, the client will use this token for auth instead
            of (automatic) secret-based auth. Ignored if the workspace
            doesn't exist.
        :param dasl_host: The URL of the DASL server. This value should
            not generally be specified unless you are testing against
            an alternative environment.
        :returns Client
        """
        try:
            return Client.new_workspace(
                admin_email,
                app_client_id,
                service_principal_id,
                service_principal_secret,
                workspace_url,
                dasl_host,
            )
        except ConflictError:
            result = Client.for_workspace(
                workspace_url, service_account_token, dasl_host
            )
            result.put_admin_config(
                AdminConfig(
                    workspace_url=workspace_url,
                    app_client_id=app_client_id,
                    service_principal_id=service_principal_id,
                    service_principal_secret=service_principal_secret,
                )
            )
            return result

    def _workspace_client(self) -> WorkspaceV1Api:
        return WorkspaceV1Api(self.auth.client())

    def _core_client(self) -> CoreV1Api:
        return CoreV1Api(self.auth.client())

    def _dbui_client(self) -> DbuiV1Api:
        return DbuiV1Api(self.auth.client())

    def _workspace(self) -> str:
        return self.auth.workspace()

    def _list_iter_paginated(
        self,
        list_func: Callable[..., Any],
        convert: Callable[[Any], T],
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Iterator[T]:
        """
        Generic helper for paginated list functions.
        """
        current_cursor = cursor
        results_so_far = 0
        while True:
            page_limit = limit - results_so_far if limit is not None else None
            if limit is not None and page_limit <= 0:
                break

            with error_handler():
                response = list_func(
                    workspace=self._workspace(),
                    cursor=current_cursor,
                    limit=page_limit,
                )

            for item in response.items:
                yield convert(item)
                results_so_far += 1

            current_cursor = (
                response.metadata.cursor if hasattr(response, "metadata") else None
            )
            if current_cursor is None:
                break

    def get_admin_config(self) -> AdminConfig:
        """
        Retrieve the AdminConfig from the DASL server. Note that the
        service principal secret will be redacted server side, so if you
        plan to make changes and issue a request using put_admin_config,
        you will need to repopulate the service_principal_secret correctly
        before passing the result back to put_admin_config.

        :returns AdminConfig
        """
        with error_handler():
            return AdminConfig.from_api_obj(
                self._workspace_client().workspace_v1_get_admin_config(
                    self._workspace()
                )
            )

    def put_admin_config(
        self,
        config: AdminConfig,
    ) -> None:
        """
        Update the AdminConfig stored in the DASL server. See the AdminConfig
        docs for details about its contents.

        :param config: AdminConfig to replace the existing. Note that the
            service principal credentials will be verified server side before
            the request is accepted.
        """
        with error_handler():
            self._workspace_client().workspace_v1_put_admin_config(
                self._workspace(),
                config.to_api_obj(),
            )

    def get_config(self) -> WorkspaceConfig:
        """
        Retrieve the WorkspaceConfig from the DASL server. The returned
        value can be updated directly and passed to put_config in order
        to make changes.
        """
        with error_handler():
            return WorkspaceConfig.from_api_obj(
                self._workspace_client().workspace_v1_get_config(self._workspace())
            )

    def put_config(
        self,
        config_in: WorkspaceConfig,
    ) -> None:
        """
        Update the WorkspaceConfig stored in the DASL server. See the
        WorkspaceConfig docs for dtails about its contents.

        :param config_in: WorkspaceConfig to replace the existing.
        :returns: WorkspaceConfig. Note that the returned value is a
            clone of config_in and may not be precisely equal to the
            originally passed value.
        """
        with error_handler():
            config = deepcopy(config_in)
            if config.metadata is None:
                config.metadata = Metadata(
                    name="config",
                    workspace=self._workspace(),
                )

            # reset the version; let the server set the version for us
            config.metadata.version = None

            self._workspace_client().workspace_v1_put_config(
                self._workspace(),
                config.to_api_obj(),
            )

    def get_datasource(self, name: str) -> DataSource:
        """
        Get the DataSource with the argument name from the DASL server. The
        returned value can be updated directly and passed to update_datasource
        in order to make changes.

        :param name: The unique name of the DataSource within this workspace
        :returns DataSource
        """
        with error_handler():
            return DataSource.from_api_obj(
                self._core_client().core_v1_get_data_source(self._workspace(), name)
            )

    def delete_datasource(self, name: str) -> None:
        """
        Delete the DataSource with the argument name from the DASL server.
        The DataSource will not necessarily be deleted immediately as the
        server will dispatch background tasks to clean up any allocated
        resources before actually deleting the resource, so it may take
        some time before its name is available for reuse.

        :param name: The unique name of the DataSource within this workspace
        """
        with error_handler():
            self._core_client().core_v1_delete_data_source(self._workspace(), name)

    def list_datasources(
        self, cursor: Optional[str] = None, limit: Optional[int] = None
    ) -> Iterator[DataSource]:
        """
        List the DataSources in this workspace. Each yielded DataSource
        contains all fields in the DataSource as if it were fetched
        using the get_datasource method.

        :param cursor: The ID of a DataSource. If specified, the results
            will contain DataSources starting (lexically) directly after
            this DataSource. If not specified, then the results will begin
            with the lexically least DataSource.
        :param limit: The maximum number of DataSources to yield. If there
            are fewer than this number of DataSources beginning directly
            after `cursor`, then all such DataSources will be yielded. If
            not specified, then all DataSources starting directly after
            `cursor` will be returned.
        :yields DataSource: One DataSource at a time in lexically
            increasing order
        """
        return self._list_iter_paginated(
            list_func=self._core_client().core_v1_list_data_sources,
            convert=DataSource.from_api_obj,
            cursor=cursor,
            limit=limit,
        )

    def create_datasource(self, name: str, ds_in: DataSource) -> DataSource:
        """
        Create a new DataSource. The chosen name must be unique for your
        workspace, and cannot refer to a DataSource that already exists
        and has not been deleted. See the documentation for delete_datasource
        as there are some caveats around name reuse.

        :param name: The unique name of this DataSource in the workspace.
        :param ds_in: The specification of the DataSource to create. See
            the documentation for the DataSource type for more details.
        :returns DataSource: Note that the returned value is a
            clone of ds_in and may not be precisely equal to the
            originally passed value.
        """
        with error_handler():
            ds = deepcopy(ds_in)
            if ds.metadata is None:
                ds.metadata = Metadata(
                    name=name,
                    workspace=self._workspace(),
                )

            result = self._core_client().core_v1_create_data_source(
                self._workspace(), ds.to_api_obj()
            )
            return DataSource.from_api_obj(result)

    def replace_datasource(self, name: str, ds_in: DataSource) -> DataSource:
        """
        Replace an existing DataSource. The name must refer to a DataSource
        that already exists in your workspace.

        :param name: The name of the existing DataSource to replace.
        :param ds_in: The specification of the DataSource taking the place
            of the existing DataSource.
        :returns DataSource: Note that the returned value is a
            clone of ds_in and may not be precisely equal to the
            originally passed value.
        """
        with error_handler():
            ds = deepcopy(ds_in)
            if ds.metadata is None:
                ds.metadata = Metadata(
                    name=name,
                    workspace=self._workspace(),
                )

            # reset the version; let the server set the version for us
            ds.metadata.version = None

            result = self._core_client().core_v1_replace_data_source(
                self._workspace(),
                name,
                ds.to_api_obj(),
            )
            return DataSource.from_api_obj(result)

    def get_rule(self, name: str) -> Rule:
        """
        Get the Rule with the argument name from the DASL server. The
        returned value can be updated directly and passed to update_rule
        in order to make changes.

        :param name: The unique name of the Rule within this workspace
        :returns Rule
        """
        with error_handler():
            return Rule.from_api_obj(
                self._core_client().core_v1_get_rule(self._workspace(), name)
            )

    def delete_rule(self, name: str) -> None:
        """
        Delete the Rule with the argument name from the DASL server.
        The Rule will not necessarily be deleted immediately as the
        server will dispatch background tasks to clean up any allocated
        resources before actually deleting the resource, so it may take
        some time before its name is available for reuse.

        :param name: The unique name of the Rule within this workspace
        """
        with error_handler():
            self._core_client().core_v1_delete_rule(self._workspace(), name)

    def list_rules(
        self, cursor: Optional[str] = None, limit: Optional[int] = None
    ) -> Iterator[Rule]:
        """
        List the Rules in this workspace. Each yielded Rule contains
        all fields in the Rule as if it were fetched using the
        get_rule method.

        :param cursor: The ID of a Rule. If specified, the results will
            contain DataSources starting (lexically) directly after this
            Rule. If not specified, then the results will begin with the
            lexically least Rule.
        :param limit: The maximum number of Rules to yield. If there are
            fewer than this number of Rules beginning directly after
            `cursor`, then all such Rules will be yielded. If not specified,
            then all Rules starting directly after `cursor` will be returned.
        :yields Rule: One Rule at a time in lexically increasing order.
        """
        return self._list_iter_paginated(
            list_func=self._core_client().core_v1_list_rules,
            convert=Rule.from_api_obj,
            cursor=cursor,
            limit=limit,
        )

    def create_rule(self, name: str, rule_in: Rule) -> Rule:
        """
        Create a new Rule. The chosen name must be unique for your
        workspace, and cannot refer to a Rule that already exists
        and has not been deleted. See the documentation for delete_rule
        as there are some caveats around name reuse.

        :param name: The unique name of this Rule in the workspace.
        :param rule_in: The specification of the Rule to create. See
            the documentation for the Rule type for more details.
        :returns Rule: Note that the returned value is a clone of
            rule_in and may not be precisely equal to the originally
            passed value.
        """
        with error_handler():
            rule = deepcopy(rule_in)
            if rule.metadata is None:
                rule.metadata = Metadata(
                    name=name,
                    workspace=self._workspace(),
                )

            result = self._core_client().core_v1_create_rule(
                self._workspace(), rule.to_api_obj()
            )
            return Rule.from_api_obj(result)

    def replace_rule(self, name: str, rule_in: Rule) -> Rule:
        """
        Replace an existing Rule. The name must refer to a Rule
        that already exists in your workspace.

        :param name: The name of the existing Rule to replace.
        :param rule_in: The specification of the Rule taking the place
            of the existing Rule.
        :returns Rule: Note that the returned value is a clone of
            rule_in and may not be precisely equal to the originally
            passed value.
        """
        with error_handler():
            rule = deepcopy(rule_in)
            if rule.metadata is None:
                rule.metadata = Metadata(
                    name=name,
                    workspace=self._workspace(),
                )

            # reset the version; let the server set the version for us
            rule.metadata.version = None

            result = self._core_client().core_v1_replace_rule(
                self._workspace(),
                name,
                rule.to_api_obj(),
            )
        return Rule.from_api_obj(result)

    def adhoc_transform(
        self,
        warehouse: str,
        request: TransformRequest,
    ) -> TransformResponse:
        """
        Run a sequence of ADHOC transforms against a SQL warehouse to
        mimic the operations performed by a datasource.

        :param warehouse: The warehouse ID to run the transforms against.
        :param request: The request containing the transforms to run.
        :return: a TransformResponse object containing the results
            after running the transforms.
        :raises: NotFoundError if the rule does not exist
        """
        with error_handler():
            return TransformResponse.from_api_obj(
                self._dbui_client().dbui_v1_transform(
                    self._workspace(),
                    warehouse,
                    request.to_api_obj(),
                )
            )

    def get_observable_events(
        self,
        warehouse: str,
        kind: str,
        value: str,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Dbui.ObservableEvents.EventsList:
        """
        Get the observable events associated with a specific field and value.

        :param warehouse: The warehouse id to perform the operation on
        :param kind: The observable kind
        :param value: The observable value
        :param cursor: A cursor to be used when paginating results
        :param limit: A limit of the number of results to return
        :returns EventsList
        """
        with error_handler():
            return Dbui.ObservableEvents.EventsList.from_api_obj(
                self._dbui_client().dbui_v1_get_observable_events(
                    self._workspace(),
                    warehouse=warehouse,
                    kind=kind,
                    value=value,
                    cursor=cursor,
                    limit=limit,
                )
            )
